# -*- coding: utf-8 -*-
"""
Created on Tue Apr 30 09:31:35 2019

@author: Karol
"""


import re
import numpy as np
from skimage import io, exposure
from skimage.util.dtype import dtype_range
from matplotlib import pyplot as plt
import os
import seaborn as sns
sns.set()

"Basic functions"
def get_L8_paths(base_path):
    substr='LC08'
    dirs=os.listdir(base_path)
    os.chdir(base_path)
    paths=[]
    for d in dirs:
        if substr in d:
            paths.append(os.path.abspath(d)+ '\\' )
    return paths
"EOF Basic functions"

"Basic Processing functions to open LANDSAT 8 files and create image and histogram"
def get_gain_bias(n):
    """
    Get band radiance gain and bias
    Input:
    n - integer in the range 1-11
    Output:
    gain - float
    bias - float
    """
    if n in range(1, 12):
        n_str = str(n)
        s_g = 'RADIANCE_MULT_BAND_' + n_str + ' = '
        s_b = 'RADIANCE_ADD_BAND_' + n_str + ' = '
        
        suf='_MTL'
        meta_file="\n".join(name for name in  os.listdir() if suf in name)
                
        f = open(meta_file, 'r+')
    
        search_str_g = '(?<=' + s_g + ').*'
        search_str_b = '(?<=' + s_b + ').*'
    
        for line in f:
            s1 = re.search(search_str_g, line)
            s2 = re.search(search_str_b, line)
            if s1:
                gain = float(s1.group(0))
            elif s2:
                bias = float(s2.group(0))
    
        f.close()
    
        return gain, bias
    else:
        print('Band number has to be in the range 1-11!')
        
def get_thermal_const(n):
    """
    Get band thermal constants K1 and K2
    Input:
    n - integer in the range 10-11
    Output:
    Ks - list of two floats
    """
    if n in range(10, 12):
        n_str = str(n)
        s = 'K._CONSTANT_BAND_' + n_str + ' = '
        search_str = '(?<=' + s + ').*'
    
        suf='_MTL'
        meta_file="\n".join(name for name in  os.listdir() if suf in name)
                
        f = open(meta_file, 'r+')
                
        Ks = []
        for line in f:
            k = re.search(search_str, line)
            if k:
                Ks.append(float(k.group(0)))
    
        f.close()
    
        return Ks
    else:
        print('Band number has to be in the range 10-11!')        

def get_gain_bias_angle(n):
    """
    Get band reflectance gain, bias, 
    and Sun elevation angle
    Input:
    n - integer in the range 1-11
    Output:
    gain - float
    bias - float
    """
    if n in range(1, 10):
        n_str = str(n)
        s_g = 'REFLECTANCE_MULT_BAND_' + n_str + ' = '
        s_b = 'REFLECTANCE_ADD_BAND_' + n_str + ' = '
    
        suf='_MTL'
        meta_file="\n".join(name for name in  os.listdir() if suf in name)
    
        f = open(meta_file, 'r+')
    
        search_str_g = '(?<=' + s_g + ').*'
        search_str_b = '(?<=' + s_b + ').*'
        search_str_a = '(?<=' + 'SUN_ELEVATION = ' + ').*'
    
        for line in f:
            s0 = re.search(search_str_a, line)
            s1 = re.search(search_str_g, line)
            s2 = re.search(search_str_b, line)
            if s0:
                angle = float(s0.group(0))
            elif s1:
                gain = float(s1.group(0))
            elif s2:
                bias = float(s2.group(0))
    
        f.close()
    
        return gain, bias, angle
    else:
        print('Band number has to be in the range 1-9!')

def get_coeffs(n):
    """
    Get band reflectance gain, bias, 
    and Sun elevation angle and sun azimuth
    Input:
    n - integer in the range 1-11
    Output:
    gain - float
    bias - float
    """
    if n in range(1, 10):
        n_str = str(n)
        s_g = 'REFLECTANCE_MULT_BAND_' + n_str + ' = '
        s_b = 'REFLECTANCE_ADD_BAND_' + n_str + ' = '
    
        suf='_MTL'
        meta_file="\n".join(name for name in  os.listdir() if suf in name)
    
        f = open(meta_file, 'r+')
    
        search_str_g = '(?<=' + s_g + ').*'
        search_str_b = '(?<=' + s_b + ').*'
        search_str_a = '(?<=' + 'SUN_ELEVATION = ' + ').*'
        search_str_az = '(?<=' + 'SUN_AZIMUTH = ' + ').*'
        search_str_d = '(?<=' + 'EARTH_SUN_DISTANCE = ' + ').*'
        
        for line in f:
            s0 = re.search(search_str_a, line)
            s1 = re.search(search_str_g, line)
            s2 = re.search(search_str_b, line)
            s3 = re.search(search_str_az, line)
            s4 = re.search(search_str_d, line)
            if s0:
                angle = float(s0.group(0))
            elif s1:
                gain = float(s1.group(0))
            elif s2:
                bias = float(s2.group(0))
            elif s3:
                azimuth = float(s3.group(0))
            elif s4:
                dist = float(s4.group(0))
        f.close()
    
        return gain, bias, angle, azimuth, dist
    else:
        print('Band number has to be in the range 1-9!')
        
def simple_DNhaze_calc(band, pixels):
    band_flat=band.flatten()
    band_flat_valid=band_flat[band_flat>0]
    band_flat_valid=np.sort(band_flat_valid)
    band_flat_valid2=band_flat_valid[:pixels]
    DN_haze=int(band_flat_valid2.mean())
    #plt.hist(r_valid)
#    plt.hist(band_flat_valid2) 
    return DN_haze
        
def image_adjust_brightness(img, limit_left, limit_right, color_map, title):
    """
    Adjust image brightness and plot the image
    Input:
    img - 2D array of uint16 type
    limit_left - integer
    limit_right - integer
    color_map - string
    title - string
    """
    img_ha = exposure.rescale_intensity(img, (limit_left, limit_right))
    
    fig = plt.figure(figsize=(10, 10))
    fig.set_facecolor('white')
    plt.imshow(img_ha, cmap=color_map)
    plt.title(title)
    plt.show()
    
    return img_ha

def image_histogram(img, title='None'):
    """
    Plot image histogram
    Input:
    img - 2D array of uint16 type
    """
    co, ce = exposure.histogram(img)
    
    fig = plt.figure(figsize=(10, 7))
    fig.set_facecolor('white')
    plt.plot(ce[1::], co[1::])
    plt.title(title)
    plt.show()
    
"End of Basic Processing functions to open LANDSAT 8 files"

