# -*- coding: utf-8 -*-
"""
Created on Tue Jun  4 09:51:46 2019

@author: Karol
"""

import basic_functions as bf

import rasterio as rio
import os
from glob import glob
import numpy as np
import fiona
from rasterio.mask import mask
from rasterio.plot import show
import matplotlib.pyplot as plt
import seaborn as sns
sns.set()

"Functions dedicated to Landsat Data preprocessing and calculations used in Main_Script"
    
def NDVI_LST_extraction(catalog_path, vector_path=None, h=0.1, mask_type='new', water_mask=True):
    """
    Get NDVI_LST two column arrays for whole triangle, wet edge and dry ege
    Input:
    catalog_path - path to landsat catalog
    
    *vector_path - path to shapefile
    *h  - subdomain width
    Output:
    NDVI_LST - numpy.array shape (n,2) type float64
    NDVI_LST_dry - numpy.array two columns type float64
    NDVI_LST_wet- numpy.array two columns type float64
    """
    os.chdir(catalog_path)

    #creating band paths list
    landsat_paths_pre = glob(
        catalog_path + "*B*.tif"
    )
    
    ##..................Vector clipping option with wrtining clipping using ready shapefile
    if vector_path!=None:
        #clipping using ready shapefile
        #creating shape lists
        vector_file = glob(
            vector_path + "\*.shp"
        )
        f_mask_prep = np.empty(0, dtype=int)
        #
        #opening clipping vector
        with fiona.open(vector_file[0], "r") as shapefile:
            features = [feature["geometry"] for feature in shapefile]
             
        for i in range(len(landsat_paths_pre)):
            #opening
            dataset= rio.open(landsat_paths_pre[i])
        
            #reading
            band = dataset.read(1)
        
            #...........subsetting
            
            #### DOS haze retrieve
            if landsat_paths_pre[i].find('_B3')!=-1:
                #opening
                dataset_Band3= rio.open(landsat_paths_pre[i])
                Band3 = dataset_Band3.read(1)
                DNhaze_B3=bf.simple_DNhaze_calc(Band3, 2)
                print('B3 DNhaze ',str(DNhaze_B3))
                b3_gain, b3_bias, angle, azimuth, dist = bf.get_coeffs(3)
                haze_B3_refl  = (b3_gain * DNhaze_B3 + b3_bias) / np.sin(angle)
                
            if landsat_paths_pre[i].find('_B4')!=-1:
                #opening
                dataset_Band4= rio.open(landsat_paths_pre[i])
                Band4 = dataset_Band4.read(1)
                DNhaze_B4=bf.simple_DNhaze_calc(Band4, 2)
                print('B4 DNhaze ',str(DNhaze_B4))
                b4_gain, b4_bias, angle, azimuth, dist = bf.get_coeffs(4)
                haze_B4_refl  = (b4_gain * DNhaze_B4 + b4_bias) / np.sin(angle)
            
            if landsat_paths_pre[i].find('_B5')!=-1:
                #opening
                dataset_Band5= rio.open(landsat_paths_pre[i])
                Band5 = dataset_Band5.read(1)
                DNhaze_B5=bf.simple_DNhaze_calc(Band5, 2)
                print('B5 DNhaze ',str(DNhaze_B5))
                b5_gain, b5_bias, angle, azimuth, dist = bf.get_coeffs(5)
                haze_B5_refl  = (b5_gain * DNhaze_B5 + b5_bias) / np.sin(angle)
            if landsat_paths_pre[i].find('_B7')!=-1:
                #opening
                dataset_Band7= rio.open(landsat_paths_pre[i])
                Band7 = dataset_Band7.read(1)
                DNhaze_B7=bf.simple_DNhaze_calc(Band7, 2)
                print('B7 DNhaze ',str(DNhaze_B7))
                b7_gain, b7_bias, angle, azimuth, dist = bf.get_coeffs(7)
                haze_B7_refl  = (b7_gain * DNhaze_B7 + b7_bias) / np.sin(angle)
            ### EOF DOS haze retrieve
            
            #cliping 
            cliped_image, cliped_transform = mask(dataset, features,
                                                                    crop=True,
                                                                    all_touched=True)
            if f_mask_prep.all():
                f_mask_prep=cliped_image
            
            out_meta = dataset.meta.copy()
            out_meta['transform']=cliped_transform
            
            #cliped catalog create if not exist
            cliped_path = catalog_path + "cliped\\"
            # print(cliped_path)
            if os.path.isdir('./cliped')==False:
                os.mkdir(cliped_path) 
            with rio.open(cliped_path + landsat_paths_pre[i].split(catalog_path)[1][:-4] + "_cliped.tif", "w", driver='GTiff',
                                         width=cliped_image.shape[2],
                                         height=cliped_image.shape[1],
                                         count=1, 
                                         dtype=out_meta['dtype'], 
                                         nodata=out_meta['nodata'],
                                         transform=out_meta['transform'], 
                                         crs=dataset.crs
    #                                     crs=pycrs.parse.from_epsg_code(dataset.crs.data['init'][5:]).to_proj4()
                                         ) as dest:
                  dest.write(cliped_image)
               
        #creating band paths list after clip
        landsat_paths_pre = glob(
                cliped_path + "*B*.tif"
                )
        
    
    ## Import Band 4 5 and 10 of LANDSAT DATA
    for band in landsat_paths_pre:
        if band.find('_B4')!=-1:
            #opening
            dataset_Band4= rio.open(band)
            if vector_path==None:
                out_meta = dataset_Band4.meta.copy()
            #reading
            Band4 = dataset_Band4.read(1)
        if band.find('_B3')!=-1:
            #opening
            dataset_Band3= rio.open(band)
            #reading
            Band3 = dataset_Band3.read(1)
            
        if band.find('_B5')!=-1:
            #opening
            dataset_Band5= rio.open(band)
            #reading
            Band5 = dataset_Band5.read(1)
            
        if band.find('_B7')!=-1:
            #opening
            dataset_Band7= rio.open(band)
            #reading
            Band7 = dataset_Band7.read(1)
            
        if band.find('_B10')!=-1:
            #opening
            dataset_Band10= rio.open(band)
            #reading
            Band10 = dataset_Band10.read(1)
        if band.find('_BQA')!=-1:
            #opening
            dataset_BandBQA= rio.open(band)
            #reading
            BandBQA = dataset_BandBQA.read(1)
            
    ## Converting Band 3,4,5 and 7 to TOA Radiance and Band 10 to temperature
    ## Processing path
            
    #Coefficients- gain, bias, sun angle
    b3_gain, b3_bias, angle = bf.get_gain_bias_angle(3)        
    b4_gain, b4_bias, angle = bf.get_gain_bias_angle(4)
    b5_gain, b5_bias, angle = bf.get_gain_bias_angle(5)
    b7_gain, b7_bias, angle = bf.get_gain_bias_angle(7)
    
    gain10, bias10 = bf.get_gain_bias(10)
    Ks = bf.get_thermal_const(10)
    #with center slolar elevation angle
    Band3_lambda_refl  = (b3_gain * Band3 + b3_bias) / np.sin(angle)
    Band4_lambda_refl  = (b4_gain * Band4 + b4_bias) / np.sin(angle)
    Band5_lambda_refl  = (b5_gain * Band5 + b5_bias) / np.sin(angle)
    Band7_lambda_refl  = (b7_gain * Band7 + b7_bias) / np.sin(angle)
    
    ### correction of haze
    Band3_lambda_refl = Band3_lambda_refl - haze_B3_refl
    Band5_lambda_refl = Band5_lambda_refl - haze_B5_refl
    Band4_lambda_refl = Band4_lambda_refl - haze_B4_refl
    Band7_lambda_refl = Band7_lambda_refl - haze_B7_refl
    ### EOF correction of haze
    
     ## masking - mask creation!!!!!!!!!!!!!!!!!!!!!!!
    #Old Data type
    if mask_type=='new':
        Fill = [1]
        C=[2800, 2804, 2808, 2812, 6896, 6900, 6904, 6908]
        CCM=[2752, 2756, 2760, 2764, 3008, 3012, 3016, 3020, 3776, 3780, 3784, 3788, 
             6848, 6852, 6856, 6860, 7104, 7108, 7112, 7116, 7872, 7876, 7880, 7884]
        CCH=[2800, 2804, 2808, 2812, 6896, 6900, 6904, 6908]
        CSH=[2976, 2980, 2984, 2988, 3008, 3012, 3016, 3020, 7072, 7076, 7080, 7084, 
             7104, 7108, 7112, 7116]
        SIH=[3744, 3748, 3752, 3756, 3776, 3780, 3784, 3788, 7840, 7844, 7848, 7852, 
             7872, 7876, 7880, 7884]
        CiCH=[6816, 6820, 6824, 6828, 6848, 6852, 6856, 6860, 6896, 6900, 6904, 6908, 
             7072, 7076, 7080, 7084, 7104, 7108, 7112, 7116, 7840, 7844, 7848, 7852, 
             7872, 7876, 7880, 7884]
        
        m_value_list=Fill+C+CCM+CCH+CSH+SIH+CiCH
        m_value_list_unique = list(set(m_value_list))
        
        BandBQA_t=BandBQA.astype(float)

        for m_val in m_value_list_unique:
           BandBQA_t[BandBQA_t == m_val]=np.NaN 
           
    elif mask_type=='old':
            
        BandBQA_t=BandBQA.astype(float)
        
        BandBQA_t[BandBQA_t == 1]=np.NaN #fill value inclusion to mask
        BandBQA_t[BandBQA_t == 31744]=np.NaN #cirrus or snow
        BandBQA_t[BandBQA_t == 28672]=np.NaN #cirrus lower
        BandBQA_t[BandBQA_t >= 45056]=np.NaN  #clouds or cirruses
            
    BandBQA_t[np.isnan(BandBQA_t)==False]=1 # other values set to 1
    
    if vector_path!=None:
        f_mask_prep=f_mask_prep.astype(float)
        f_mask_prep[f_mask_prep!=0]=1
        f_mask_prep[f_mask_prep==0]=np.NaN
        f_mask_prep=f_mask_prep[0]
        BandBQA_t=BandBQA_t*f_mask_prep
    
    Mask=BandBQA_t
    
    # Allow division by zero
    np.seterr(divide='ignore', invalid='ignore')
    NDVI = (Band5_lambda_refl - Band4_lambda_refl) / (Band5_lambda_refl + Band4_lambda_refl)
    MNDWI = (Band3_lambda_refl - Band7_lambda_refl) / (Band3_lambda_refl + Band7_lambda_refl)
    
    ################# masking 
    NDVI_masked=NDVI*Mask
    MNDWI_masked=MNDWI*Mask
    #Temperature
    
    ##calculation on arrays
    L  = gain10 * Band10 + bias10 # Convert raw pixel readouts to spectral radiance
    TEMP = Ks[1] / np.log(Ks[0] / L + 1)
    
    #################################   MAIN PART TVDI calculation 
    ## TVDI calculation on arrays
    #Coefficients

    NDVI_v=np.nanpercentile(NDVI_masked, 99.99)
    NDVI_s=np.nanpercentile(NDVI_masked, 0.01)
    eps_veg_l=0.985
    eps_soil_l=0.93
    C_l=0.005
    
    #preparation
    #Fractional cover
    PV = ((NDVI - NDVI_s)/(NDVI_v - NDVI_s))**2
    
    #Epsilon
    EPS = eps_veg_l*PV + eps_soil_l*(1-PV)+C_l
    #temperature
    LST=TEMP/(EPS**0.25)
    LST_masked=LST*Mask
    
    #....Excluding pixels qualified as water
    #water mak creation
    if water_mask == True:
        print('Water masking procedure')
        water_mask_MNDWI=MNDWI_masked <=0.3
        water_mask_MNDWI= water_mask_MNDWI.astype(float)
        show(water_mask_MNDWI)
        water_mask_MNDWI[water_mask_MNDWI == 0]=np.NaN
        
        #masking
        NDVI_masked=NDVI_masked*water_mask_MNDWI
        LST_masked=LST_masked*water_mask_MNDWI
    else:
        water_mask_MNDWI=1
        
    #################################   MAIN PART TVDI calculation
    
    
    NDVI_c=np.reshape(NDVI_masked,(NDVI_masked.size,1))
    LST_c=np.reshape(LST_masked,(LST_masked.size,1))
    
    #excluding nan values
    NDVI_cc=NDVI_c[np.isnan(NDVI_c)==False]
    LST_cc=LST_c[np.isnan(NDVI_c)==False]
    
    NDVI_c=NDVI_cc[np.isnan(LST_cc)==False]
    LST_c=LST_cc[np.isnan(LST_cc)==False]
    
    #crating array to calculation
    NDVI_LST=np.array([NDVI_c, LST_c])
    
    #Transposition
    NDVI_LST=NDVI_LST.T
    NDVI_LST=NDVI_LST
     
    # Sorting by NDVI values
    NDVI_LST=NDVI_LST[NDVI_LST[:,0].argsort()]
    
    
    #Extracting points
    start_NDVI = NDVI_LST[0,0]
    end_NDVI=NDVI_LST[NDVI_LST.shape[0]-1,0]
    sd_NDVI_begin = start_NDVI
    i=1
    
    #setting empy arrays
    NDVI_LST_dry=np.array([])
    NDVI_LST_wet=np.array([])
    
    while sd_NDVI_begin <= end_NDVI:
        sd_i=NDVI_LST[NDVI_LST[:,0]>=sd_NDVI_begin]
        sd_i=sd_i[sd_i[:,0]<sd_NDVI_begin+h]
        #Subdomain length
        sd_i_count=len(sd_i)
#        print(sd_i_count)
        if sd_i_count!=0:
    #        #Ts_max with coreesponding NDVI_min in pairs !!!
            
            NDVI_LST_dry=np.append(NDVI_LST_dry, sd_i[sd_i[:,1]==max(sd_i[:,1])])
#            print(sd_i[sd_i[:,1]==max(sd_i[:,1])])
            NDVI_LST_dry=NDVI_LST_dry.reshape(int(NDVI_LST_dry.size/2),2)
        
            #Ts_min with coreesponding NDVI_max in pairs
            NDVI_LST_wet=np.append(NDVI_LST_wet, sd_i[sd_i[:,1]==min(sd_i[:,1])])
#            print(sd_i[sd_i[:,1]==min(sd_i[:,1])])
            NDVI_LST_wet=NDVI_LST_wet.reshape(int(NDVI_LST_wet.size/2),2)
            
        i=i+1
        sd_NDVI_begin=sd_NDVI_begin+h
    
    print('End of TVDI edges pixel ectraction')    
    return NDVI_LST_dry, NDVI_LST_wet, NDVI_LST, NDVI, LST, Mask, out_meta, water_mask_MNDWI

def TVDI_calculate (NDVI_LST_dry, NDVI_LST_wet, NDVI_LST, h, catalog_path):

    """
    Function calculate TVDI and qTVDI. It saves pictures, txt files and returns coefficients
    Get NDVI_LST_dry, NDVI_LST_wet, NDVI_LST 
    Input:
    NDVI_LST - two column array, result from NDVI_LST_extraction
    NDVI_LST - numpy.array shape (n,2) type float64
    NDVI_LST_dry - numpy.array two columns type float64
    NDVI_LST_wet- numpy.array two columns type float64
    
    Output:
    edge_coeffs
    """
    #.............Coeficient computing section
    # Linear wet from whole NDVI range
    coeff_wet = np.polyfit( NDVI_LST_wet[:,0], NDVI_LST_wet[:,1], 1 )
    # Linear dry from NDVI range 0.2-0.9 
    sub_NDVI_LST_dry=NDVI_LST_dry[NDVI_LST_dry[:,0]>=0.2]
    sub_NDVI_LST_dry=sub_NDVI_LST_dry[sub_NDVI_LST_dry[:,0]<=0.9]
    sub_coeff_dry = np.polyfit( sub_NDVI_LST_dry[:,0], sub_NDVI_LST_dry[:,1], 1 )

 
    #QUadratic from whole NDVI range
    coeff_dry_2 = np.polyfit( NDVI_LST_dry[:,0], NDVI_LST_dry[:,1], 2 )
    coeff_wet_2 = np.polyfit( NDVI_LST_wet[:,0], NDVI_LST_wet[:,1], 2 )
    	
    print('Linear Coeffs for Dry from 0.2 to 0.9')
    print(sub_coeff_dry)
    print('Linear Coeffs for Wet')
    print(coeff_wet)
    print('Quadratic Coeffs for Dry')
    print(coeff_dry_2)
    print('Quadratic Coeffs for  Wet')
    print(coeff_wet_2)
    #..... End of Coeficient computing section
    
    #............ Ploting section
    
    ##Creating analysis directory to save textfiles and figures
    os.chdir(catalog_path)
    analysis_path = catalog_path + "analysis\\"
    if os.path.isdir('./analysis')==False:
        os.mkdir(analysis_path) 
    
    
    # ploting aproximation 
    poly_wet=np.poly1d(coeff_wet)
    poly_dry_sub=np.poly1d(sub_coeff_dry)
    
    poly_dry2=np.poly1d(coeff_dry_2)
    poly_wet2=np.poly1d(coeff_wet_2)
    x=np.linspace(-1,1,100)
    
    # ***** TVDI classic edge plot
    plt.figure(1)
    #plt.subplot(211)
    plt.plot(NDVI_LST_dry[:,0],NDVI_LST_dry[:,1], 'ro')
    #plt.subplot(212)
    plt.plot(NDVI_LST_wet[:,0],NDVI_LST_wet[:,1], 'bo')
    plt.plot(x, poly_dry_sub(x), 'r', label="Dry edge")
    plt.plot(x, poly_wet(x), 'b', label="Wet edge")
    
    plt.title('Classic TVDI edges with dry egde fitting range NDVI 0.2 - 0.9, from ' +  catalog_path[-24:-16])
    plt.ylabel('LST [K]',  rotation=90, fontsize=14)
    plt.xlabel('NDVI [-]', fontsize=14)
    plt.legend(loc=1, ncol=1, borderaxespad=0.)
    plt.savefig(analysis_path + catalog_path[-24:-16]+ '_classic_TVDI_h' + str(h) + '.jpg', dpi=300)
    plt.show()
    
    # ***** end of TVDI classic edge plot
    
    # *******TVDI quadratic edge plot
    plt.figure(2)
    plt.plot(NDVI_LST_dry[:,0],NDVI_LST_dry[:,1], 'ro')
    plt.plot(NDVI_LST_wet[:,0],NDVI_LST_wet[:,1], 'bo')
    plt.plot(x, poly_dry2(x), 'r--', label="Quadratic Dry edge")
    plt.plot(x, poly_wet2(x), 'b--', label="Quadratic Wet edge")  
    
    plt.ylabel('LST [K]',  rotation=90, fontsize=14)
    plt.xlabel('NDVI [-]', fontsize=14)
    plt.legend(loc=1, ncol=1, borderaxespad=0.)
    plt.title('Quadratic TVDI edges from ' +  catalog_path[-24:-16])
    plt.savefig(analysis_path + catalog_path[-24:-16]+'_quadratic_TVDI_h' + str(h) + '.jpg', dpi=300)
    plt.show()
       
    #******* end of TVDI quadratic edge plot
    
    # whole triangle, takes very long time (couple of minutes) because of 
    #the NDVI_LST matrix size
    plt.figure(3)
     
    plt.plot(NDVI_LST[:,0],NDVI_LST[:,1], 'bo',alpha=0.05)
    plt.plot(x, poly_dry_sub(x), 'r', label="Dry edge")
    plt.plot(x, poly_wet(x), 'b', label="Wet edge")
    plt.plot(x, poly_dry2(x), 'r--', label="Quadratic Dry edge")
    plt.plot(x, poly_wet2(x), 'b--', label="Quadratic Wet edge")
    
    plt.ylabel('LST [K]',  rotation=90, fontsize=14)
    plt.xlabel('NDVI [-]', fontsize=14)
    plt.legend(loc=1, ncol=1, borderaxespad=0.)
    plt.suptitle('Scatterplot with all pixels with Classic and Quadratic Edges')
    plt.title(catalog_path[-24:-16])
    plt.savefig(analysis_path + 'Triangle_TVDI_h' + str(h) +'_'+ catalog_path[-24:-16]+ '.jpg', dpi=300)
    plt.show()
    
    #...........End of Ploting section
   
    ##......... Saving files .........
    
    np.savetxt(analysis_path + 'NDVI_LST_dry_h' + str(h) + '_'+ catalog_path[-24:-16]+ '.txt',NDVI_LST_dry)
    np.savetxt(analysis_path + 'NDVI_LST_wet_h' + str(h) + '_'+ catalog_path[-24:-16]+'.txt',NDVI_LST_wet)
    # to all pixels, uncomment only if neccesery, it takes large ammount of memory about 120mB.
    np.savetxt(analysis_path + 'NDVI_LST_h' + str(h) + '_'+ catalog_path[-24:-16]+'.txt',NDVI_LST)
    ##........end of Saving files.......
    return sub_coeff_dry, coeff_wet, coeff_dry_2, coeff_wet_2
      
def qTVDI_image (coeff_dry_2, coeff_wet_2, NDVI, LST, Mask,  catalog_path, out_meta, h, water_mask_MNDWI,  show_org=True, show_masked=True):

    """
    Function creates qTVDI image. 

    """
    #..........Creating qTVDI image

    #Quadratic coeff_dry_2, coeff_wet_2
    TVDI_quadratic=(LST-(coeff_wet_2[2]+coeff_wet_2[1]*NDVI+coeff_wet_2[0]*NDVI**2))/((coeff_dry_2[2]+coeff_dry_2[1]*NDVI+coeff_dry_2[0]*NDVI**2)-(coeff_wet_2[2]+coeff_wet_2[1]*NDVI+coeff_wet_2[0]*NDVI**2))
    TVDI_quadratic_masked = TVDI_quadratic*Mask
    TVDI_quadratic_masked=TVDI_quadratic_masked*water_mask_MNDWI
    #............ Optional show part
    if show_org==True:
        #showing and histogram of classic without masking
        show(TVDI_quadratic, title='TVDI qadratic without masking')
        bf.image_histogram(TVDI_quadratic, title='Histogram of TVDI quadratic without masking')
        
    if show_masked==True:
        #showing and histogram of classic with masking
        show(TVDI_quadratic_masked, title='TVDI qadratic with masking')
        #Next line excludes nans in order to draw histogram
        TVDI_quadratic_masked_wn=TVDI_quadratic_masked[np.isnan(TVDI_quadratic_masked)==False]
        bf.image_histogram(TVDI_quadratic_masked_wn, title='Histogram of TVDI quadratic with masking')
        
    # Saving part
    os.chdir(catalog_path)
    TVDI_path = catalog_path + "TVDI\\"
    if os.path.isdir('./TVDI')==False:
        os.mkdir(TVDI_path)
    
    #Creating 3d numy array from 2d - as in case of original tiffs
    TVDI_quadratic_3d=np.hstack([[TVDI_quadratic_masked]])
    
    with rio.open(TVDI_path + catalog_path.split('\\')[-2] + "_TVDI_quadratic_"+ str(h) +".tif", "w", driver='GTiff',
                                         width=TVDI_quadratic_3d.shape[2],
                                         height=TVDI_quadratic_3d.shape[1],
                                         count=TVDI_quadratic_3d.shape[0], 
                                         dtype=TVDI_quadratic_masked.dtype, 
                                         nodata=out_meta['nodata'],
                                         transform=out_meta['transform'], 
                                         crs=out_meta['crs']
                                         ) as dest:
                  dest.write(TVDI_quadratic_3d)
    
def TVDI_image (sub_coeff_dry, coeff_wet, NDVI, LST, Mask, catalog_path, out_meta, h, water_mask_MNDWI,  show_org=True, show_masked=True):

    """
    Function creates qTVDI image
    """
    #..........Creating TVDI image
    #Classic sub_coeff_dry, coeff_wet
    TVDI_classic=(LST-(coeff_wet[1]+coeff_wet[0]*NDVI))/((sub_coeff_dry[1]+sub_coeff_dry[0]*NDVI)-(coeff_wet[1]+coeff_wet[0]*NDVI))
    TVDI_classic_masked = TVDI_classic*Mask
    TVDI_classic_masked=TVDI_classic_masked*water_mask_MNDWI
    #............ Optional show part of histogra
    if show_org==True:
        #showing and histogram of classic without masking
        show(TVDI_classic,title='TVDI classic without masking')
        bf.image_histogram(TVDI_classic, title='Histogram of TVDI classic without masking')
        
    if show_masked==True:
        #showing and histogram of classic with masking
        show(TVDI_classic_masked, title='TVDI classic with masking')
        #Next line excludes nans in order to draw histogram
        TVDI_classic_masked_wn=TVDI_classic_masked[np.isnan(TVDI_classic_masked)==False]
        bf.image_histogram(TVDI_classic_masked_wn, title='Histogram of TVDI classic with masking')
        
    # Saving part
    os.chdir(catalog_path)
    TVDI_path = catalog_path + "TVDI\\"
    if os.path.isdir('./TVDI')==False:
        os.mkdir(TVDI_path) 
        
    #Creating 3d numy array from 2d - as in case of original tiffs
    TVDI_classic_3d=np.hstack([[TVDI_classic_masked]])
    with rio.open(TVDI_path + catalog_path.split('\\')[-2] + "_TVDI_classic_"+ str(h) +".tif", "w", driver='GTiff',
                                         width=TVDI_classic_3d.shape[2],
                                         height=TVDI_classic_3d.shape[1],
                                         count=TVDI_classic_3d.shape[0], 
                                         dtype=TVDI_classic_masked.dtype, 
                                         nodata=out_meta['nodata'],
                                         transform=out_meta['transform'], 
                                         crs=out_meta['crs']
                                         ) as dest:
                     dest.write(TVDI_classic_3d)
                  
def Other_ind_calc(catalog_path, vector_path=None, mask_type='new', water_mask=True):
    """
    Other pixel by pixel index calculation and saving
    """
    os.chdir(catalog_path)

    #creating band paths list
    landsat_paths_pre = glob(
        catalog_path + "*B*.tif"
    )
    
    ##..................Vector clipping option with wrtining clipping using ready shapefile
    if vector_path!=None:
        #clipping using ready shapefile
        #creating shape lists
        vector_file = glob(
            vector_path + "\*.shp"
        )
        f_mask_prep = np.empty(0, dtype=int)
        #
        #opening clipping vector
        with fiona.open(vector_file[0], "r") as shapefile:
            features = [feature["geometry"] for feature in shapefile]
            
        for i in range(len(landsat_paths_pre)):
            #opening
            dataset= rio.open(landsat_paths_pre[i])       
            #reading
            band = dataset.read(1)
        
            #...........subsetting
            
            #### DOS haze retrieve
            if landsat_paths_pre[i].find('_B2')!=-1: #blue
                #opening
                dataset_Band2= rio.open(landsat_paths_pre[i])
                Band2 = dataset_Band2.read(1)
                DNhaze_B2=bf.simple_DNhaze_calc(Band2, 2)
                print('B2 DNhaze ',str(DNhaze_B2))
                b2_gain, b2_bias, angle, azimuth, dist = bf.get_coeffs(2)
                haze_B2_refl  = (b2_gain * DNhaze_B2 + b2_bias) / np.sin(angle)
            if landsat_paths_pre[i].find('_B3')!=-1: #green
                #opening
                dataset_Band3= rio.open(landsat_paths_pre[i])
                Band3 = dataset_Band3.read(1)
                DNhaze_B3=bf.simple_DNhaze_calc(Band3, 2)
                print('B3 DNhaze ',str(DNhaze_B3))
                b3_gain, b3_bias, angle, azimuth, dist = bf.get_coeffs(3)
                haze_B3_refl  = (b3_gain * DNhaze_B3 + b3_bias) / np.sin(angle)   
            if landsat_paths_pre[i].find('_B4')!=-1: #red
                #opening
                dataset_Band4= rio.open(landsat_paths_pre[i])
                Band4 = dataset_Band4.read(1)
                DNhaze_B4=bf.simple_DNhaze_calc(Band4, 2)
                print('B4 DNhaze ',str(DNhaze_B4))
                b4_gain, b4_bias, angle, azimuth, dist = bf.get_coeffs(4)
                haze_B4_refl  = (b4_gain * DNhaze_B4 + b4_bias) / np.sin(angle)
            if landsat_paths_pre[i].find('_B5')!=-1: #nir
                #opening
                dataset_Band5= rio.open(landsat_paths_pre[i])
                Band5 = dataset_Band5.read(1)
                DNhaze_B5=bf.simple_DNhaze_calc(Band5, 2)
                print('B5 DNhaze ',str(DNhaze_B5))
                b5_gain, b5_bias, angle, azimuth, dist = bf.get_coeffs(5)
                haze_B5_refl  = (b5_gain * DNhaze_B5 + b5_bias) / np.sin(angle)
            if landsat_paths_pre[i].find('_B6')!=-1: #swir
                #opening
                dataset_Band6= rio.open(landsat_paths_pre[i])
                Band6 = dataset_Band6.read(1)
                DNhaze_B6=bf.simple_DNhaze_calc(Band6, 2)
                print('B6 DNhaze ',str(DNhaze_B6))
                b6_gain, b6_bias, angle, azimuth, dist = bf.get_coeffs(6)
                haze_B6_refl  = (b6_gain * DNhaze_B6 + b6_bias) / np.sin(angle)
            if landsat_paths_pre[i].find('_B7')!=-1: #Mwir
                #opening
                dataset_Band7= rio.open(landsat_paths_pre[i])
                Band7 = dataset_Band7.read(1)
                DNhaze_B7=bf.simple_DNhaze_calc(Band7, 2)
                print('B7 DNhaze ',str(DNhaze_B7))
                b7_gain, b7_bias, angle, azimuth, dist = bf.get_coeffs(7)
                haze_B7_refl  = (b7_gain * DNhaze_B7 + b7_bias) / np.sin(angle)
                
            ### EOF DOS haze retrieve
            
            #cliping 
            cliped_image, cliped_transform = mask(dataset, features,
                                                                    crop=True,
                                                                    all_touched=True)
            
            if f_mask_prep.all():
                f_mask_prep=cliped_image
            
            out_meta = dataset.meta.copy()
            out_meta['transform']=cliped_transform
            
            #cliped catalog create if not exist
            cliped_path = catalog_path + "cliped\\"
            # print(cliped_path)
            if os.path.isdir('./cliped')==False:
                os.mkdir(cliped_path) 
            with rio.open(cliped_path + landsat_paths_pre[i].split(catalog_path)[1][:-4] + "_cliped.tif", "w", driver='GTiff',
                                         width=cliped_image.shape[2],
                                         height=cliped_image.shape[1],
                                         count=1, 
                                         dtype=out_meta['dtype'], 
                                         nodata=out_meta['nodata'],
                                         transform=out_meta['transform'], 
                                         crs=dataset.crs) as dest:
                  dest.write(cliped_image)
               
        #creating band paths list after clip
        landsat_paths_pre = glob(
                cliped_path + "*B*.tif"
                )
    ## Import Band 2 3 4 5 6 7 and 10 of LANDSAT DATA
    for band in landsat_paths_pre:
        
        if band.find('_B2')!=-1:
            #opening
            dataset_Band2= rio.open(band)
            #reading
            Band2 = dataset_Band2.read(1)
        if band.find('_B3')!=-1:
            #opening
            dataset_Band3= rio.open(band)
            #reading
            Band3 = dataset_Band3.read(1)
            
        if band.find('_B4')!=-1:
            #opening
            dataset_Band4= rio.open(band)
            if vector_path==None:
                out_meta = dataset_Band4.meta.copy()
             
            #dataset = rio.open(catalog_path)
            #reading
            Band4 = dataset_Band4.read(1)
        if band.find('_B5')!=-1:
            #opening
            dataset_Band5= rio.open(band)
            #reading
            Band5 = dataset_Band5.read(1)
        if band.find('_B6')!=-1: #swir
            #opening
            dataset_Band6= rio.open(band)
            #reading
            Band6 = dataset_Band6.read(1)
        if band.find('_B7')!=-1: #swir
            #opening
            dataset_Band7= rio.open(band)
            #reading
            Band7 = dataset_Band7.read(1)
        if band.find('_B10')!=-1:
            #opening
            dataset_Band10= rio.open(band)
            #reading
            Band10 = dataset_Band10.read(1)
        if band.find('_BQA')!=-1:
            #opening
            dataset_BandBQA= rio.open(band)
            #reading
            BandBQA = dataset_BandBQA.read(1)
            
    ## Processing path
            
    #Coefficients- gain, bias, sun angle
    
    b2_gain, b2_bias, angle = bf.get_gain_bias_angle(2)
    b3_gain, b3_bias, angle = bf.get_gain_bias_angle(3)         
    b4_gain, b4_bias, angle = bf.get_gain_bias_angle(4)
    b5_gain, b5_bias, angle = bf.get_gain_bias_angle(5)
    b6_gain, b6_bias, angle = bf.get_gain_bias_angle(6)
    b7_gain, b7_bias, angle = bf.get_gain_bias_angle(7)
    
    gain10, bias10 = bf.get_gain_bias(10)
    Ks = bf.get_thermal_const(10)
    #with center slolar elevation angle
    
    Band2_lambda_refl  = (b2_gain * Band2 + b2_bias) / np.sin(angle)
    Band3_lambda_refl  = (b3_gain * Band3 + b3_bias) / np.sin(angle)
    Band4_lambda_refl  = (b4_gain * Band4 + b4_bias) / np.sin(angle)
    Band5_lambda_refl  = (b5_gain * Band5 + b5_bias) / np.sin(angle)
    Band6_lambda_refl  = (b6_gain * Band6 + b6_bias) / np.sin(angle)
    Band7_lambda_refl  = (b7_gain * Band7 + b7_bias) / np.sin(angle)
    
    ### correction of haze
    Band2_lambda_refl = Band2_lambda_refl - haze_B2_refl
    Band3_lambda_refl = Band3_lambda_refl - haze_B3_refl
    Band4_lambda_refl = Band4_lambda_refl - haze_B4_refl
    Band5_lambda_refl = Band5_lambda_refl - haze_B5_refl
    Band6_lambda_refl = Band6_lambda_refl - haze_B6_refl
    Band7_lambda_refl = Band7_lambda_refl - haze_B7_refl
       
    ### EOF correction of haze
     ## masking - mask creation!!!!!!!!!!!!!!!!!!!!!!!
    #Old Data type
    if mask_type=='new':
        Fill = [1]
        C=[2800, 2804, 2808, 2812, 6896, 6900, 6904, 6908]
        CCM=[2752, 2756, 2760, 2764, 3008, 3012, 3016, 3020, 3776, 3780, 3784, 3788, 
             6848, 6852, 6856, 6860, 7104, 7108, 7112, 7116, 7872, 7876, 7880, 7884]
        CCH=[2800, 2804, 2808, 2812, 6896, 6900, 6904, 6908]
        CSH=[2976, 2980, 2984, 2988, 3008, 3012, 3016, 3020, 7072, 7076, 7080, 7084, 
             7104, 7108, 7112, 7116]
        SIH=[3744, 3748, 3752, 3756, 3776, 3780, 3784, 3788, 7840, 7844, 7848, 7852, 
             7872, 7876, 7880, 7884]
        CiCH=[6816, 6820, 6824, 6828, 6848, 6852, 6856, 6860, 6896, 6900, 6904, 6908, 
             7072, 7076, 7080, 7084, 7104, 7108, 7112, 7116, 7840, 7844, 7848, 7852, 
             7872, 7876, 7880, 7884]
        
        m_value_list=Fill+C+CCM+CCH+CSH+SIH+CiCH
        m_value_list_unique = list(set(m_value_list))
        
        BandBQA_t=BandBQA.astype(float)

        for m_val in m_value_list_unique:
           BandBQA_t[BandBQA_t == m_val]=np.NaN 
           
    elif mask_type=='old':
            
        BandBQA_t=BandBQA.astype(float)
        
        BandBQA_t[BandBQA_t == 1]=np.NaN #fill value inclusion to mask
        BandBQA_t[BandBQA_t == 31744]=np.NaN #cirrus or snow
        BandBQA_t[BandBQA_t == 28672]=np.NaN #cirrus lower
        BandBQA_t[BandBQA_t >= 45056]=np.NaN  #clouds or cirruses
            
    BandBQA_t[np.isnan(BandBQA_t)==False]=1 # other values set to 1
    
    if vector_path!=None:
        f_mask_prep=f_mask_prep.astype(float)
        f_mask_prep[f_mask_prep!=0]=1
        f_mask_prep[f_mask_prep==0]=np.NaN
        f_mask_prep=f_mask_prep[0]
        BandBQA_t=BandBQA_t*f_mask_prep
    
    Mask=BandBQA_t
    
    # Allow division by zero
    np.seterr(divide='ignore', invalid='ignore')
    #Indices
    NDVI = (Band5_lambda_refl - Band4_lambda_refl) / (Band5_lambda_refl + Band4_lambda_refl)
    EVI = 2.5 * ((Band5_lambda_refl  - Band4_lambda_refl) / (Band5_lambda_refl + 6 * Band4_lambda_refl - 7.5 * Band2_lambda_refl + 1))
    NDBI = (Band6_lambda_refl - Band5_lambda_refl) / (Band6_lambda_refl + Band5_lambda_refl)
    NDWI = (Band3_lambda_refl - Band5_lambda_refl) / (Band3_lambda_refl + Band5_lambda_refl)
    MNDWI = (Band3_lambda_refl - Band7_lambda_refl) / (Band3_lambda_refl + Band7_lambda_refl)
    NDMI = (Band5_lambda_refl - Band6_lambda_refl) / (Band5_lambda_refl + Band6_lambda_refl)
    ########### Masking 
    NDVI_masked=NDVI*Mask
    EVI_masked=EVI*Mask
    NDBI_masked=NDBI*Mask
    NDWI_masked=NDWI*Mask
    MNDWI_masked=MNDWI*Mask
    NDMI_masked=NDMI*Mask
    #Temperature
    
    ##calculation on arrays
    L  = gain10 * Band10 + bias10 # Convert raw pixel readouts to spectral radiance
    TEMP = Ks[1] / np.log(Ks[0] / L + 1)
    
    NDVI_v=np.nanpercentile(NDVI_masked, 99.99)
    print(NDVI_v)
    NDVI_s=np.nanpercentile(NDVI_masked, 0.01) 
    print(NDVI_s)
    eps_veg_l=0.985
    eps_soil_l=0.93
    C_l=0.005
    
    #preparation
    #Fractional cover
    PV = ((NDVI - NDVI_s)/(NDVI_v - NDVI_s))**2
    
    #Epsilon
    EPS = eps_veg_l*PV + eps_soil_l*(1-PV)+C_l
    #temperature
    LST=TEMP/(EPS**0.25)
    
    LST_masked=LST*Mask
    
    #....Excluding pixels qualified as water
    #water mak creation
    if water_mask == True:
        print('Water masking procedure')
        water_mask_MNDWI=MNDWI_masked <=0.3
        water_mask_MNDWI= water_mask_MNDWI.astype(float)
        water_Bodies=water_mask_MNDWI.copy()
        water_mask_MNDWI[water_mask_MNDWI == 0]=np.NaN
        
        water_Bodies[water_Bodies == 0]=2
        water_Bodies[water_Bodies == 1]=np.NaN
        water_Bodies[water_Bodies == 2]=1
        
        water_Bodies=water_Bodies*Mask
        water_Bodies[np.isnan(water_Bodies)]=0#linijka podmieniajaca
        #masking
        NDVI_masked=NDVI_masked*water_mask_MNDWI
        EVI_masked=EVI_masked*water_mask_MNDWI
        NDBI_masked=NDBI_masked*water_mask_MNDWI
        NDMI_masked=NDMI_masked*water_mask_MNDWI
        LST_masked=LST_masked*water_mask_MNDWI

    #Saving other indices to catalogs
    # Saving part
    os.chdir(catalog_path)
    other_path = catalog_path + "other\\"
    if os.path.isdir('./other')==False:
        os.mkdir(other_path) 
    
    #Creating 3d numy array from 2d - as in case of original tiffs
    LST_masked_3d=np.hstack([[LST_masked]])
    with rio.open(other_path + catalog_path.split('\\')[-2] + "_LST.tif", "w", driver='GTiff',
                                         width=LST_masked_3d.shape[2],
                                         height=LST_masked_3d.shape[1],
                                         count=LST_masked_3d.shape[0], 
                                         dtype=LST_masked.dtype, 
                                         nodata=out_meta['nodata'],
                                         transform=out_meta['transform'], 
                                         crs=out_meta['crs']
                                         ) as dest:
                     dest.write(LST_masked_3d)
                     
    NDVI_masked_3d=np.hstack([[NDVI_masked]])
    with rio.open(other_path + catalog_path.split('\\')[-2] + "_NDVI.tif", "w", driver='GTiff',
                                         width=NDVI_masked_3d.shape[2],
                                         height=NDVI_masked_3d.shape[1],
                                         count=NDVI_masked_3d.shape[0], 
                                         dtype=NDVI_masked.dtype, 
                                         nodata=out_meta['nodata'],
                                         transform=out_meta['transform'], 
                                         crs=out_meta['crs']
                                         ) as dest:
                     dest.write(NDVI_masked_3d)
                     
    EVI_masked_3d=np.hstack([[EVI_masked]])
    with rio.open(other_path + catalog_path.split('\\')[-2] + "_EVI.tif", "w", driver='GTiff',
                                         width=EVI_masked_3d.shape[2],
                                         height=EVI_masked_3d.shape[1],
                                         count=EVI_masked_3d.shape[0], 
                                         dtype=EVI_masked.dtype, 
                                         nodata=out_meta['nodata'],
                                         transform=out_meta['transform'], 
                                         crs=out_meta['crs']
                                         ) as dest:
                     dest.write(EVI_masked_3d)
    NDBI_masked_3d=np.hstack([[NDBI_masked]])
    with rio.open(other_path + catalog_path.split('\\')[-2] + "_NDBI.tif", "w", driver='GTiff',
                                         width=NDBI_masked_3d.shape[2],
                                         height=NDBI_masked_3d.shape[1],
                                         count=NDBI_masked_3d.shape[0], 
                                         dtype=NDBI_masked.dtype, 
                                         nodata=out_meta['nodata'],
                                         transform=out_meta['transform'], 
                                         crs=out_meta['crs']
                                         ) as dest:
                     dest.write(NDBI_masked_3d)
    NDWI_masked_3d=np.hstack([[NDWI_masked]])
    with rio.open(other_path + catalog_path.split('\\')[-2] + "_NDWI.tif", "w", driver='GTiff',
                                         width=NDWI_masked_3d.shape[2],
                                         height=NDWI_masked_3d.shape[1],
                                         count=NDWI_masked_3d.shape[0], 
                                         dtype=NDWI_masked.dtype, 
                                         nodata=out_meta['nodata'],
                                         transform=out_meta['transform'], 
                                         crs=out_meta['crs']
                                         ) as dest:
                     dest.write(NDWI_masked_3d)
    MNDWI_masked_3d=np.hstack([[MNDWI_masked]])
    with rio.open(other_path + catalog_path.split('\\')[-2] + "_MNDWI.tif", "w", driver='GTiff',
                                         width=MNDWI_masked_3d.shape[2],
                                         height=MNDWI_masked_3d.shape[1],
                                         count=MNDWI_masked_3d.shape[0], 
                                         dtype=MNDWI_masked.dtype, 
                                         nodata=out_meta['nodata'],
                                         transform=out_meta['transform'], 
                                         crs=out_meta['crs']
                                         ) as dest:
                     dest.write(MNDWI_masked_3d)
                     
    NDMI_masked_3d=np.hstack([[NDMI_masked]])
    with rio.open(other_path + catalog_path.split('\\')[-2] + "_NDMI.tif", "w", driver='GTiff',
                                         width=NDMI_masked_3d.shape[2],
                                         height=NDMI_masked_3d.shape[1],
                                         count=NDMI_masked_3d.shape[0], 
                                         dtype=NDMI_masked.dtype, 
                                         nodata=out_meta['nodata'],
                                         transform=out_meta['transform'], 
                                         crs=out_meta['crs']
                                         ) as dest:
                     dest.write(NDMI_masked_3d)
                     
    
    if water_mask == True:
        water_mask_MNDWI_3d=np.hstack([[water_mask_MNDWI]])
        with rio.open(other_path + catalog_path.split('\\')[-2] + "_water_mask.tif", "w", driver='GTiff',
                                             width=NDMI_masked_3d.shape[2],
                                             height=NDMI_masked_3d.shape[1],
                                             count=NDMI_masked_3d.shape[0], 
                                             dtype=NDMI_masked_3d.dtype, 
                                             nodata=out_meta['nodata'],
                                             transform=out_meta['transform'], 
                                             crs=out_meta['crs']
                                             ) as dest:
                         dest.write(water_mask_MNDWI_3d)
        water_Bodies_3d=np.hstack([[water_Bodies]])
        with rio.open(other_path + catalog_path.split('\\')[-2] + "_water_Bodies.tif", "w", driver='GTiff',
                                             width=water_Bodies_3d.shape[2],
                                             height=water_Bodies_3d.shape[1],
                                             count=water_Bodies_3d.shape[0], 
                                             dtype=water_Bodies_3d.dtype, 
                                             nodata=out_meta['nodata'],
                                             transform=out_meta['transform'], 
                                             crs=out_meta['crs']
                                             ) as dest:
                         dest.write(water_Bodies_3d)
                         
    Mask_3d=np.hstack([[Mask]])
    with rio.open(other_path + catalog_path.split('\\')[-2] + "_Mask.tif", "w", driver='GTiff',
                                         width=Mask_3d.shape[2],
                                         height=Mask_3d.shape[1],
                                         count=Mask_3d.shape[0], 
                                         dtype=Mask_3d.dtype, 
                                         nodata=out_meta['nodata'],
                                         transform=out_meta['transform'], 
                                         crs=out_meta['crs']
                                         ) as dest:
                     dest.write(Mask_3d)

    return LST_masked, NDVI_masked, EVI_masked, NDBI_masked, NDWI_masked, MNDWI_masked 
 
